// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "list.h"
#include "wmpf/hardware/camera.h"
#include "wmpf/voip.h"
#include "wmpf/wmpf.h"

#define DEMO_USE_H265

typedef struct {
  struct wx_camera_stream_config config;
  const struct wx_camera_stream_listener* listener;
  void* user_data;
  struct wx_camera_stream* stream;
  struct listnode slist;
} CAMERA_STREAM;

static list_declare(camera_stream_list);
extern wx_voip_session_status_t voip_status;

static pthread_mutex_t camera_mutex;

static void dump_cameracfg(const struct wx_camera_stream_config* config) {
  printf("\tconfig->format = %x\n", config->format);
  printf("\tconfig->pixel_format = %x\n", config->pixel_format);
}

static void camera_device_close(struct wx_device* device) {}

static wx_error_t camera_device_getparameter(
    const struct wx_metadata* metadata,
    uint32_t key,
    struct wx_metadata_entry* value_out) {
  uint32_t size[2] = {640, 480};

  if (!value_out) {
    return WXERROR_INVALID_ARGUMENT;
  }

  switch (key) {
    case WX_CAMERA_DEVICE_METADATA_FACING:
      wx_metadata_entry_set_u32(value_out, key, WX_CAMERA_FACING_FRONT);
      return WXERROR_OK;

    case WX_CAMERA_DEVICE_METADATA_FRAMERATE:
      wx_metadata_entry_set_u32(value_out, key, 30);
      return WXERROR_OK;

    case WX_CAMERA_DEVICE_METADATA_BITRATE:
      wx_metadata_entry_set_u32(value_out, key, 200000);
      return WXERROR_OK;

    case WX_CAMERA_DEVICE_METADATA_SIZE:
      wx_metadata_entry_set_u32_array(value_out, key, size, 2);
      return WXERROR_OK;

    default:
      return WXERROR_NOT_FOUND;
  }
}

static wx_error_t camera_device_setparameter(
    struct wx_metadata* metadata,
    uint32_t key,
    const struct wx_metadata_entry* value) {
  if (!value) {
    return WXERROR_INVALID_ARGUMENT;
  }

  switch (value->key) {
    case WX_CAMERA_DEVICE_METADATA_FRAMERATE:
      if (value->type != WX_METADATA_ENTRY_TYPE_U32)
        return WXERROR_INVALID_ARGUMENT;

      printf("camear fps = %d\n", value->data.u32);
      return WXERROR_OK;

    case WX_CAMERA_DEVICE_METADATA_SIZE:
      if (value->type != WX_METADATA_ENTRY_TYPE_U32_ARRAY || value->count != 2)
        return WXERROR_INVALID_ARGUMENT;

      printf("camera size %d %d\n", value->data.u32_array[0],
             value->data.u32_array[1]);
      return WXERROR_OK;

    default:
      return WXERROR_NOT_FOUND;
  }
}

static bool camera_supports(struct wx_camera_device* dev,
                            const struct wx_camera_stream_config* config) {
  return true;
}

static void camera_stream_close(struct wx_camera_stream* stream) {
  printf("camera_stream_close\n");
  dump_cameracfg(&stream->config);

  pthread_mutex_lock(&camera_mutex);

  struct listnode* node;
  list_for_each(node, &camera_stream_list) {
    CAMERA_STREAM* _stream = node_to_item(node, CAMERA_STREAM, slist);

    if (_stream->stream == stream) {
      list_remove(&_stream->slist);
      free(_stream->stream);
      free(_stream);
      break;
    }
  }

  pthread_mutex_unlock(&camera_mutex);

  return;
}

static struct wx_camera_stream* alloc_camera_stream(void) {
  struct wx_camera_stream* stream =
      (struct wx_camera_stream*)calloc(1, sizeof(struct wx_camera_stream));

  if (stream) {
    stream->common.size = sizeof(struct wx_camera_stream);
    stream->common.tag = WX_CAMERA_STREAM_TAG;
    stream->common.version = 0;

    stream->config.format = WX_VIDEO_FORMAT_H264;
    stream->config.pixel_format = WX_VIDEO_PIXEL_FORMAT_YUV420;

    stream->close = camera_stream_close;
  }

  return stream;
}

static wx_error_t camera_open_stream(
    struct wx_camera_device* dev,
    const struct wx_camera_stream_config* config,
    const struct wx_camera_stream_listener* listener,
    void* user_data,
    struct wx_camera_stream** stream_out) {
  printf("%s user_data=%p\n", __FUNCTION__, user_data);
  dump_cameracfg(config);

#ifdef DEMO_USE_H265
  if (config->format == WX_VIDEO_FORMAT_H265) {
    printf("WX_VIDEO_FORMAT_H265\n");
  }
  if (config->format == WX_VIDEO_FORMAT_H264) {
    printf("WX_VIDEO_FORMAT_H264\n");
    return WXERROR_UNIMPLEMENTED;
  }
#else
  if (config->format == WX_VIDEO_FORMAT_H265) {
    printf("WX_VIDEO_FORMAT_H265\n");
    return WXERROR_UNIMPLEMENTED;
  }
  if (config->format == WX_VIDEO_FORMAT_H264) {
    printf("WX_VIDEO_FORMAT_H264\n");
  }
#endif

  CAMERA_STREAM* stream = (CAMERA_STREAM*)calloc(1, sizeof(CAMERA_STREAM));

  if (stream) {
    struct wx_camera_stream* _stream = alloc_camera_stream();

    if (_stream) {
      memcpy(&stream->config, config, sizeof(struct wx_camera_stream_config));
      stream->listener = listener;
      stream->stream = _stream;
      stream->user_data = user_data;

      list_add_tail(&camera_stream_list, &stream->slist);
      *stream_out = _stream;

      return WXERROR_OK;
    }

    free(stream);
  }

  return WXERROR_RESOURCE_EXHAUSTED;
}

static struct wx_camera_device camera_device = {
    .common =
        {
            .common =
                {
                    .tag = WX_CAMERA_DEVICE_TAG,
                    .size = sizeof(struct wx_camera_device),
                    .version = 0,
                },
            .priv = NULL,
            .close = camera_device_close,
        },
    .metadata =
        {
            .get_parameter = camera_device_getparameter,
            .set_parameter = camera_device_setparameter,
            .priv = &camera_device,
        },
    .supports = camera_supports,
    .open_stream = camera_open_stream,
};

static wx_error_t camera_get_number_of_devices(struct wx_camera_module* module,
                                               size_t* num_devices_out) {
  *num_devices_out = 1;
  return WXERROR_OK;
}

static wx_error_t camera_get_device_info(
    struct wx_camera_module* module,
    size_t index,
    struct wx_camera_device_info* device_info) {
  printf("%s index = %zu\n", __FUNCTION__, index);
  return WXERROR_OK;
}

static wx_error_t camera_open(struct wx_camera_module* module,
                              const char* id,
                              struct wx_camera_device** device_out) {
  printf("%s %s\n", __FUNCTION__, id);

  if (!device_out) {
    return WXERROR_INVALID_ARGUMENT;
  }

  if (!strcmp(id, WX_CAMERA_DEVICE_PRIMARY)) {
    *device_out = (struct wx_camera_device*)&camera_device;
    return WXERROR_OK;
  }

  return WXERROR_NOT_FOUND;
}

struct wx_camera_module camera_module = {
    .common =
        {
            .common =
                {
                    .tag = WX_CAMERA_MODULE_TAG,
                    .size = sizeof(struct wx_camera_module),
                    .version = 0,
                },
            .id = WX_CAMERA_MODULE_ID,
            .set_on_devices_changed = NULL,
        },
    .get_number_of_devices = camera_get_number_of_devices,
    .get_device_info = camera_get_device_info,
    .open = camera_open,
};

static pthread_t pid = 0;
static int thread_exit = 0;

extern int h264_init(const char* path);
int h264_readnalu(uint8_t* p, size_t* size, int* type);

static void* thread_camera(void* data) {
#ifdef DEMO_USE_H265
  if (h264_init("./h265.h265") < 0) {
    printf("open h265.h265 fail\n");
    return NULL;
  }
#else
  if (h264_init("./h264.h264") < 0) {
    printf("open h264.h264 fail\n");
    return NULL;
  }
#endif

  uint8_t* buffer = (uint8_t*)malloc(1024 * 1024);

  if (!buffer) {
    printf("thread_camera nomem\n");
    return NULL;
  }

  size_t sz = 0;
  int tp = 0;
  int timeout = 0;

  while (!thread_exit) {
    if (voip_status == WX_VOIP_SESSION_TALKING) {
      /*
       * 这里延时1秒后再从文件第一桢开始发送
       * 真实的设备流不需要这样
       */
      if (timeout++ < 1) {
        usleep(1000 * 1000);
        continue;
      }

      /*
       * 这里使用h264文件里的数据，会使用 h264_readnalu 来进行 h264 流的初步解析
       * 开发者应该使用真实的 h264 流数据。
       */
      int rd = h264_readnalu(buffer, &sz, &tp);

      if (rd <= 0) {
        printf("ReadNalu error 1\n");
        usleep(1000 * 1000);
        continue;
      }

      pthread_mutex_lock(&camera_mutex);

      struct listnode* node;
      list_for_each(node, &camera_stream_list) {
        CAMERA_STREAM* stream;

        stream = node_to_item(node, CAMERA_STREAM, slist);

#ifdef DEMO_USE_H265
        if (stream->config.format == WX_VIDEO_FORMAT_H265) {
          struct timespec ts = {0};

          stream->listener->data(stream->stream, stream->user_data, buffer, sz,
                                 0, 0, WX_VIDEO_ROTATION_0, ts);
        }
#else
        if (stream->config.format == WX_VIDEO_FORMAT_H264) {
          struct timespec ts = {0};

          stream->listener->data(stream->stream, stream->user_data, buffer, sz,
                                 0, 0, WX_VIDEO_ROTATION_0, ts);
        }
#endif
      }

      pthread_mutex_unlock(&camera_mutex);
    }

    usleep(1000 * 66);
  }

  return NULL;
}

void stop_camera_thread(void) {
  thread_exit = 1;
  pthread_join(pid, NULL);
}

void start_camera_thread(void) {
  pthread_create(&pid, NULL, thread_camera, NULL);
}
