// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "wmpf/crypto.h"
#include "wmpf/hardware/audio.h"
#include "wmpf/hardware/camera.h"
#include "wmpf/hardware/video.h"
#include "wmpf/veridev.h"
#include "wmpf/voip.h"
#include "wmpf/wmpf.h"

extern struct wx_audio_module audio_module;
extern struct wx_camera_module camera_module;
extern struct wx_crypto_module crypto_module;
extern struct wx_video_module_t video_module;

void start_audio_thread(void);
void stop_audio_thread(void);
void start_camera_thread(void);
void stop_camera_thread(void);

wx_voip_session_status_t voip_status;

static char device_id[256] = {0};
static char device_sig[512] = {0};
static char model_id[256] = {0};
static char appid[256] = {0};
static char host_appid[256] = {0};
static char sn_ticket[256] = {0};
static int product_id = 0;

static char openid[256] = {0};
static char roomid[256] = {0};

wx_error_t hal_get_module(const char* id, struct wx_module** module_out) {
  if (!strcmp(id, WX_AUDIO_MODULE_ID)) {
    *module_out = (struct wx_module*)&audio_module;
    return WXERROR_OK;
  }
  if (!strcmp(id, WX_CAMERA_MODULE_ID)) {
    *module_out = (struct wx_module*)&camera_module;
    return WXERROR_OK;
  }
  if (!strcmp(id, WX_CRYPTO_MODULE_ID)) {
    *module_out = (struct wx_module*)&crypto_module;
    return WXERROR_OK;
  }
  if (!strcmp(id, WX_VIDEO_MODULE_ID)) {
    *module_out = (struct wx_module*)&video_module;
    return WXERROR_OK;
  }

  return WXERROR_UNIMPLEMENTED;
}

static void status_change(wx_voip_session_t session,
                          void* user_data,
                          wx_voip_session_status_t s) {
  voip_status = s;

  printf("voip_status = %d\n", voip_status);

  return;
}

struct wx_voip_session_listener voip_session_listener = {
    .common =
        {
            .tag = WX_VOIP_SESSION_LISTENER_TAG,
            .size = sizeof(struct wx_voip_session_listener),
            .version = 0,
        },
    .status = status_change,
};

static wx_error_t voip_init_config(void) {
  wx_init_config_t config = {0};

  config.common.tag = WX_INIT_CONFIG_TAG;
  config.common.size = sizeof(wx_init_config_t);
  config.common.version = 0;

  // 以下参数需要参考文档注释传入正确的值，均不可填空或 0.
  config.device_id = device_id;
  config.device_signature = device_sig;
  config.device_signature_version = 1;
  config.data_dir = "/data";
  config.host_appid = host_appid;
  config.model_id = model_id;
  config.product_id = product_id;
  config.wxa_appid = appid;
  config.rpmb_device = NULL;
  
  // h265_only 可以让对方只发265流 
  // config.h265_only = 1;
  // video_landscape 可以让对方发landscape流
  // config.video_landscape = 1;
  // subscribe_video_length 可以订阅对方发过来流的分辨率长值
  // config.subscribe_video_length = 640;

  wx_operation_t op = wx_init(&config, hal_get_module);
  printf("op = %p\n", op);

  return wx_operation_wait(op, 0);
}

int main(int argc, char** argv) {

  if (argc < 9) {
    printf(
        "voipdemo product_id device_id device_sig model_id appid host_appid "
        "sn_ticket openid [roomid]\n");
    return -1;
  }

  product_id = atoi(argv[1]);
  strcpy(device_id, argv[2]);
  strcpy(device_sig, argv[3]);
  strcpy(model_id, argv[4]);
  strcpy(appid, argv[5]);
  strcpy(host_appid, argv[6]);
  strcpy(sn_ticket, argv[7]);
  strcpy(openid, argv[8]);

  if (argc == 10) {
    strcpy(roomid, argv[9]);
  }

  /* 初始化 wmpf 框架 ,
   * 使用了同步等待的方式，开发者可根据需求修改为异步+回调的方式 */
  wx_error_t init_ret = voip_init_config();
  if (init_ret != WXERROR_OK) {
    printf("voip init fail, %d\n", init_ret);
    return EXIT_FAILURE;
  }

  bool is_registered = false;
  wx_error_t is_reg_ret = wx_device_is_registered(&is_registered);
  if (is_reg_ret != WXERROR_OK) {
    printf("wx_device_is_registered fail, %d\n", is_reg_ret);
    return EXIT_FAILURE;
  }

  if (!is_registered) {
    printf("Device is not registered. Do registration now.\n");

    // sn_ticket
    // 获取参考：https://developers.weixin.qq.com/miniprogram/dev/OpenApiDoc/hardware-device/getSnTicket.html

    wx_operation_t op = wx_device_register(sn_ticket);
    wx_error_t ret = wx_operation_wait(op, 0);
    if (ret != WXERROR_OK) {
      printf("wx_device_register fail, %d\n", ret);
      return EXIT_FAILURE;
    }
  }

  if (strstr(argv[0], "hangup") && strlen(roomid) != 0) {

    printf("wx_voip_listener_hangup\n");
    wx_operation_t op = wx_voip_listener_hangup(roomid,
                                      WX_VOIP_HANGUP_REASON_MANUAL);
    wx_error_t ret = wx_operation_wait(op, 0);
    if (ret != WXERROR_OK) {
      printf("wx_voip_listener_hangup wait fail\n");
    }
    wx_stop();
    return 0;  
  }

  /* 创建一个 voip session */
  wx_voip_session_t session;

  /* 设备身份 */
  wx_voip_member_t device_member = {
      "device_name",
      device_id,
      WX_VOIP_MEMBER_CAMERA_STATUS_OPEN,
  };

  /* 小程序身份 */
  wx_voip_member_t wechat_member = {
      "name",
      openid,
      WX_VOIP_MEMBER_CAMERA_STATUS_OPEN,
  };

  start_audio_thread();
  start_camera_thread();

  wx_voip_session_config_t session_config = {0, 0, 0};

  if (wx_voip_session_new(WX_VOIP_SESSION_VIDEO, WX_VOIP_SESSION_SCENE_IOT,
                          &voip_session_listener, NULL, &device_member,
                          &session_config, &session) != WXERROR_OK) {
    printf("session new fail\n");
    wx_stop();
    return -1;
  }

  /* 有 roomid 参数时，演示反向呼叫加入房间 */
  if (argc == 10) {
    wx_operation_t op = wx_voip_listener_join(session, roomid);
    wx_error_t ret = wx_operation_wait(op, 0);
    if (ret != WXERROR_OK) {
      printf("wx_voip_session_call wait fail\n");
    }

    getchar();

    op = wx_voip_session_hangup(session, WX_VOIP_HANGUP_REASON_MANUAL);
    ret = wx_operation_wait(op, 0);
    if (ret != WXERROR_OK) {
      printf("wx_voip_session_hangup wait fail %d\n", ret);
    }
    wx_voip_session_destroy(session);
    wx_stop();

    return 0;
  }


  /* 发起呼叫
   * callee.name: 微信用户昵称（显示在通话页面中，可以填房间号之类的名称）
   * callee.id: 微信用户的 OpenID.
   */

  while (1) {
    wx_operation_t op = wx_voip_session_call(session, &wechat_member);
    wx_error_t ret = wx_operation_wait(op, 0);
    if (ret != WXERROR_OK) {
      printf("wx_voip_session_call wait fail\n");
    }

    printf("wx_voip_session_get_roomid3 = %s\n", wx_voip_session_get_roomid(session));

    getchar();

    op = wx_voip_session_hangup(session, WX_VOIP_HANGUP_REASON_MANUAL);
    ret = wx_operation_wait(op, 0);
    if (ret != WXERROR_OK) {
      printf("wx_voip_session_hangup wait fail %d\n", ret);
    }

    getchar();
  }

  wx_voip_session_destroy(session);
  wx_stop();

  return 0;
}
