// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include "wmpf/types.h"

WX_BEGIN_DECLS

#define WX_FIXED_FRACTIONAL_BITS (6)
#define WX_FIXED_POINT_DENOMINATOR (1 << WX_FIXED_FRACTIONAL_BITS)

/**
 * @brief 定点浮点数
 *
 * 为了防止在运算过程中出现精度误差, 我们使用定点浮点数来表示可能有小数的数字.
 *
 * 请注意，不要将 int、double 等数据类型直接传入到 wx_fixed_t 类型的变量,
 * 需要使用 wx_fixed_from_int, wx_fixed_from_float, wx_fixed_from_double
 * 函数来初始化 wx_fixed_t 变量.
 */
typedef int wx_fixed_t;

/**
 * @brief 定点浮点数转 32 位有符号整型
 *
 * @param fixed 定点浮点数
 * @return 将定点浮点数向下去整后的整数
 */
WX_ALLOW_UNUSED_TYPE static int wx_fixed_to_int(wx_fixed_t fixed) {
  return fixed / WX_FIXED_POINT_DENOMINATOR;
}

/**
 * @brief 定点浮点数转单精度浮点数
 *
 * @param fixed 定点浮点数
 * @return 转换后的单精度浮点数
 */
WX_ALLOW_UNUSED_TYPE static float wx_fixed_to_float(wx_fixed_t fixed) {
  return (float)fixed / WX_FIXED_POINT_DENOMINATOR;
}

/**
 * @brief 定点浮点数转双精度浮点数
 *
 * @param fixed 定点浮点数
 * @return 转换后的双精度浮点数
 */
WX_ALLOW_UNUSED_TYPE static double wx_fixed_to_double(wx_fixed_t fixed) {
  return (double)fixed / WX_FIXED_POINT_DENOMINATOR;
}

/**
 * @brief 32 位有符号整数转定点浮点数
 * 要转换的整数范围必须在 -33554432~33554431 范围内.
 *
 * @param i 要转换的 32 位有符号整数
 * @return 转换后的定点浮点数
 */
WX_ALLOW_UNUSED_TYPE static wx_fixed_t wx_fixed_from_int(int i) {
  return i * WX_FIXED_POINT_DENOMINATOR;
}

/**
 * @brief 单精度浮点数转定点浮点数
 * 要转换的浮点数范围必须在 -33554432~33554431 范围内.
 *
 * @param f 要转换的单精度浮点数
 * @return 转换后的定点浮点数
 */
WX_ALLOW_UNUSED_TYPE static wx_fixed_t wx_fixed_from_float(float f) {
  return (wx_fixed_t)(f * WX_FIXED_POINT_DENOMINATOR);
}

/**
 * @brief 双精度浮点数转定点浮点数
 * 要转换的浮点数范围必须在 -33554432~33554431 范围内.
 *
 * @param d 要转换的双精度浮点数
 * @return 转换后的定点浮点数
 */
WX_ALLOW_UNUSED_TYPE static wx_fixed_t wx_fixed_from_double(double d) {
  return (wx_fixed_t)(d * WX_FIXED_POINT_DENOMINATOR);
}

WX_END_DECLS
