// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include "wmpf/hardware/device.h"
#include "wmpf/hardware/video.h"
#include "wmpf/module.h"
#include "wmpf/types.h"

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

WX_BEGIN_DECLS

#define WX_CAMERA_MODULE_ID "camera"

#define WX_CAMERA_DEVICE_PRIMARY "primary"

typedef enum wx_camera_device_metadata_type {
  /**
   * @brief 摄像头方向
   * 类型: u32
   * 值范围: wx_camera_facing_t
   */
  WX_CAMERA_DEVICE_METADATA_FACING = 0,

  /**
   * @brief FPS
   * 类型: u32
   */
  WX_CAMERA_DEVICE_METADATA_FRAMERATE = 1,

  /**
   * @brief 比特率
   * 类型: u32
   */
  WX_CAMERA_DEVICE_METADATA_BITRATE = 2,

  /**
   * @brief 录像像素宽高
   * 类型: u32_array
   * 值范围: count 必须为 2
   */
  WX_CAMERA_DEVICE_METADATA_SIZE = 3,
} wx_camera_device_metadata_type_t;

/**
 * @brief 摄像头方向
 *
 */
typedef enum wx_camera_facing {
  /**
   * @brief 后置摄像头
   *
   */
  WX_CAMERA_FACING_BACK = 0,
  /**
   * @brief 前置摄像头
   *
   */
  WX_CAMERA_FACING_FRONT = 1,
} wx_camera_facing_t;

typedef struct wx_camera_stream_config {
  /**
   * @brief 摄像头输出数据格式
   * 该项必填.
   */
  enum wx_video_format format;

  /**
   * @brief 摄像头视频输入流的像素格式
   * 当视频流数据格式为 WX_VIDEO_FORMAT_RAW 时，该项必填 WX_VIDEO_PIXEL_FORMAT;
   * 档视频流数据格式为其他编码流时，该项可填 WX_VIDEO_PIXEL_UNKNOWN.
   */
  enum wx_video_pixel_format pixel_format;

} wx_camera_stream_config_t;

#define WX_CAMERA_STREAM_TAG 0x30003

typedef struct wx_camera_stream_encoding_config {
  /**
   * @brief 期望设备摄像头像素宽度
   *
   */
  size_t width;

  /**
   * @brief 期望设备摄像头像素高度
   *
   */
  size_t height;

  /**
   * @brief 期望码率
   *
   */
  size_t bitrate;

  /**
   * @brief 期望 FPS
   *
   */
  uint8_t framerate;

  /**
   * @brief 期望 I 帧间隔
   *
   */
  uint8_t i_frame_period;
} wx_camera_stream_encoding_config_t;

typedef struct wx_camera_stream {
  struct wx_struct common;

  /**
   * @brief 该图像流的参数
   *
   */
  wx_camera_stream_config_t config;

  /**
   * @brief 请求更新摄像头录制参数
   * 属于微信 VoIP 服务端质量管理策略, 该请求可以被忽略.
   *
   * 厂商可以不实现该函数.
   */
  wx_error_t (*update)(struct wx_camera_stream* stream,
                       const struct wx_camera_stream_encoding_config*);

  /**
   * @brief 建议立刻创建 I 帧
   * VoIP 后台 QoS 可能会建议设备端立刻发送一个 I 帧以保证视频通话质量.
   * 一般是刚拨通电话时, 需要摄像头视频流立刻编码一个 PPS+SPS+I 帧,
   * 以便对方能立刻显示出画面, 否则对方需要等待到下一个 I 帧才能看到画面.
   *
   * 厂商可以不实现该函数.
   */
  wx_error_t (*make_i_frame)(struct wx_camera_stream* stream);

  void (*close)(struct wx_camera_stream* stream);
} wx_camera_stream_t;

#define WX_CAMERA_STREAM_LISTENER_TAG 0x30004

typedef struct wx_camera_stream_listener {
  struct wx_struct common;

  /**
   * @brief 新捕获的视频帧
   * @param buffer 视频帧数据
   * @param bytes 视频帧数据长度
   * @param width 视频帧像素高度
   * @param height 视频帧像素宽度
   * @param rotation 视频帧画面旋转角度，参考值：0, 90, 180, 270.
   * @param capture_time 视频帧捕获时间戳
   */
  void (*data)(struct wx_camera_stream* stream,
               void* user_data,
               const void* buffer,
               size_t bytes,
               size_t width,
               size_t height,
               enum wx_video_rotation rotation,
               struct timespec capture_time);

  void (*error)(struct wx_camera_stream* stream,
                void* user_data,
                wx_error_t error);
} wx_camera_stream_listener_t;

#define WX_CAMERA_DEVICE_INFO_TAG 0x30005

typedef struct wx_camera_device_info {
  struct wx_struct common;

  char* id;
  enum wx_camera_facing facing;
} wx_camera_device_info_t;

#define WX_CAMERA_DEVICE_TAG 0x30006

/**
 * @brief 摄像头设备
 * 摄像头设备被打开后，厂商实现可以实现为独占摄像头设备，并准备数据输出。在 WMPF
 * 不再需要摄像头数据后， WMPF 将会关闭摄像头设备。厂商此时可以释放摄像头。
 * 这里的设计类似 Android Camera API 的 openCamera 独占行为。
 */
typedef struct wx_camera_device {
  struct wx_device common;

  struct wx_metadata metadata;

  bool (*supports)(struct wx_camera_device* dev,
                   const struct wx_camera_stream_config* config);

  /**
   * @brief 创建摄像头设备的视频输入流
   *
   * 需要注意, WMPF 常常需要摄像头同时输出 H.264/H.265 编码后的视频流 (即
   * format==WX_VIDEO_FORMAT_H264 或者 format==WX_VIDEO_FORMAT_H265), 以及
   * format==WX_VIDEO_FORMAT_RAW && pixel_format==WX_VIDEO_PIXEL_FORMAT_XRGB
   * 的图像流.
   * 厂商在实现摄像头图像数据捕获时，必须处理同一个摄像头同时创建多个包含不同
   * 目标编码及目标像素格式的视频输入流,
   * 必要时需要在后台自行调用硬件编码器编码视频流.
   *
   * @param config 视频录制配置
   * @param listener 视频流回调, 需要保证生命周期长于返回的视频流
   * @param user_data 视频流回调用户数据, 需要保证生命周期长于返回的视频流
   * @return 操作是否成功
   *   - WXERROR_INVALID_ARGUMENT:
   *     请求创建图像流的长宽不受传感器支持
   */
  wx_error_t (*open_stream)(struct wx_camera_device* dev,
                            const struct wx_camera_stream_config* config,
                            const struct wx_camera_stream_listener* listener,
                            void* user_data,
                            struct wx_camera_stream** stream_out);

} wx_camera_device_t;

#define WX_CAMERA_MODULE_TAG 0x30000

/**
 * @brief 摄像头模块抽象类
 *
 */
typedef struct wx_camera_module {
  struct wx_module common;

  /**
   * @brief 获取摄像头设备数量
   * @param num_devices_out 保存当前时刻摄像头设备数量
   * @return 获取结果
   *   - WXERROR_INVALID_ARGUMENT: num_devices_out 为空
   */
  wx_error_t (*get_number_of_devices)(struct wx_camera_module* module,
                                      size_t* num_devices_out);

  /**
   * @brief 获取指定摄像头设备信息
   * @param index
   */
  wx_error_t (*get_device_info)(struct wx_camera_module* module,
                                size_t index,
                                struct wx_camera_device_info* device_info);

  /**
   * @brief 打开指定设备
   *
   * 如果当前设备包含至少一个摄像头, 厂商就必须在 id 传入
   * WX_CAMERA_DEVICE_PRIMARY 时能返回给 VoIP 使用的那个摄像头.
   *
   * @param id 设备 id, 若传入 WX_CAMERA_DEVICE_PRIMARY,
   * 则必须返回一个可用的摄像头设备 (若系统内有摄像头设备).
   * @param device_out 保存要打开的设备
   * @return 打开结果
   *   - WXERROR_INVALID_ARGUMENT: device_out 为空
   *   - WXERROR_NOT_FOUND: 找不到指定 id 的设备
   */
  wx_error_t (*open)(struct wx_camera_module* module,
                     const char* id,
                     struct wx_camera_device** device_out);
} wx_camera_module_t;

WX_END_DECLS
