// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include "wmpf/hardware/device.h"
#include "wmpf/module.h"
#include "wmpf/types.h"

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

WX_BEGIN_DECLS

#define WX_VIDEO_MODULE_ID "video"

struct wx_video_decoder;
struct wx_video_encoder;
struct wx_video_decoder;

typedef enum wx_video_format_profile {
  WX_VIDEO_FORMAT_H264_PROFILE_BASELINE = 0,
  WX_VIDEO_FORMAT_H264_PROFILE_MAIN = 1,
  WX_VIDEO_FORMAT_H264_PROFILE_EXTENDED = 2,
  WX_VIDEO_FORMAT_H264_PROFILE_HIGH = 3,
  WX_VIDEO_FORMAT_H264_PROFILE_HIGH_10 = 4,
  WX_VIDEO_FORMAT_H264_PROFILE_HIGH_422 = 5,
  WX_VIDEO_FORMAT_H264_PROFILE_HIGH_444_PREDICTIVE = 6,
  WX_VIDEO_FORMAT_H264_PROFILE_SCALABLE_BASELINE = 7,
  WX_VIDEO_FORMAT_H264_PROFILE_SCALABLE_HIGH = 8,
  WX_VIDEO_FORMAT_H264_PROFILE_STEREO_HIGH = 9,
  WX_VIDEO_FORMAT_H264_PROFILE_MULTIVIEW_HIGH = 10,
} wx_video_format_profile_t;

typedef enum wx_video_format {
  WX_VIDEO_FORMAT_UNKNOWN = 0x00000000UL,

  // clang-format off
  WX_VIDEO_FORMAT_RAW  = 0x00100000UL,
  WX_VIDEO_FORMAT_H264 = 0x00200000UL,
  WX_VIDEO_FORMAT_H265 = 0x00300000UL,
  // clang-format on

  WX_VIDEO_FORMAT_MASK = 0xFFF00000UL,
} wx_video_format_t;

typedef enum wx_video_pixel_format {
  WX_VIDEO_PIXEL_FORMAT_UNKNOWN = 0,

  WX_VIDEO_PIXEL_FORMAT_YUV420 = 1,

  /**
   * @brief 一个像素 4 个字节 XRGB8888.
   * 这种视频流通常用于屏幕渲染, 因为屏幕渲染直出 XRGB8888 的图像数据.
   * 第一个字节=X (该字节不读取, 可填任意值, 建议填 0)
   * 第二个字节=R (该字节存储红色通道数据 0~255)
   * 第三个字节=G (该字节存储绿色通道数据 0~255)
   * 第四个字节=B (该字节存储蓝色通道数据 0~255)
   */
  WX_VIDEO_PIXEL_FORMAT_XRGB = 2,
  WX_VIDEO_PIXEL_FORMAT_YUY2 = 3,
  WX_VIDEO_PIXEL_FORMAT_NV21 = 4,
  WX_VIDEO_PIXEL_FORMAT_NV12 = 5,
} wx_video_pixel_format_t;

/**
 * @brief
 *
 */
typedef enum wx_video_rotation {
  WX_VIDEO_ROTATION_0 = 0,
  WX_VIDEO_ROTATION_90 = 90,
  WX_VIDEO_ROTATION_180 = 180,
  WX_VIDEO_ROTATION_270 = 270,
} wx_video_rotation_t;

#define WX_VIDEO_ENCODER_CONFIG_TAG 0x40001

/**
 * @brief 视频流编码器配置
 *
 */
typedef struct wx_video_encoder_config {
  /**
   * @brief 结构体信息，使用时应当检查内容是否匹配，避免传入了错误的结构体
   *
   * tag=WX_VIDEO_ENCODER_TAG
   * size=sizeof(struct wx_video_encoder_config)
   * version=0
   */
  struct wx_struct common;

  /**
   * @brief 视频流像素宽度
   */
  uint32_t width;

  /**
   * @brief 视频流像素高度
   */
  uint32_t height;

  /**
   * @brief 要输出的视频流的编码
   */
  enum wx_video_format format;

  /**
   * @brief 视频帧的像素格式
   *
   */
  enum wx_video_pixel_format pixel_format;
} wx_video_encoder_config_t;

#define WX_VIDEO_ENCODER_LISTENER_TAG 0x40002

/**
 * @brief 视频流编码器回调
 *
 */
typedef struct wx_video_encoder_listener {
  struct wx_struct common;

  /**
   * @brief 视频流编码器编码后数据回调
   * 视频流编码器在接收到足够多的要执行编码的视频帧并得到数据后，应当尽快调用该函数返回视频编码后数据
   * 不要求一次性将得到的所有数据通过该函数返回，可以在一组数据编码过程中多次调用本函数返回数据片段.
   *
   * @param buffer 编码后的视频流数据
   * @param buffer_size 数据字节数
   */
  wx_error_t (*data)(struct wx_video_encoder* encoder,
                     void* user_data,
                     const void* buffer,
                     size_t buffer_size);

  /**
   * @brief 通知视频流编码器在编码时出现的错误
   * @param encoder 出现错误的视频流编码器
   * @param user_data 通过 create_encoder 传入的 user_data
   * @param status 错误信息
   */
  void (*error)(struct wx_video_encoder* encoder,
                void* user_data,
                wx_error_t status);
} wx_video_encoder_listener_t;

#define WX_VIDEO_ENCODER_TAG 0x40003

/**
 * @brief 视频流编码器
 *
 */
typedef struct wx_video_encoder {
  struct wx_device common;

  /**
   * @brief 向视频编码器提供一个视频帧数据
   * 视频流编码器在接收到足够多的视频帧数据后，应当调用 listener->data
   * 回调通知编码后的数据.
   *
   * @param buffer 要编码的视频帧数据
   * @param buffer_size 数据字节数
   */
  wx_error_t (*write)(struct wx_video_encoder* encoder,
                      const void* buffer,
                      size_t buffer_size);
} wx_video_encoder_t;

#define WX_VIDEO_DECODER_CONFIG_TAG 0x40004

/**
 * @brief 视频流解码器配置
 *
 */
typedef struct wx_video_decoder_config {
  struct wx_struct common;

  /**
   * @brief 视频流像素宽度
   */
  uint32_t width;

  /**
   * @brief 视频流像素高度
   */
  uint32_t height;

  /**
   * @brief 要解码的视频流的原编码格式
   */
  enum wx_video_format format;

  /**
   * @brief 要解码的视频流的像素格式
   * 如果视频流帧中已经包含了像素格式, 该项可填 WX_VIDEO_PIXEL_FORMAT_UNKNOWN,
   * 由解码器自行判断.
   */
  enum wx_video_pixel_format pixel_format;

  /**
   * @brief 解码后视频帧的像素格式
   *
   * 如果该项填 WX_VIDEO_PIXEL_FORMAT_UNKNOWN,
   * 表示取原视频流当前视频帧当前的像素格式.
   *
   * 注意，解码器还需在 pixel_format 与 target_pixel_format 不一致时,
   * 自行变换像素格式. 如在播放视频时, 视频解码得到 YUV422 格式, 但 UI 渲染
   * 时需要 XRGB888 格式, 此时就需要视频解码器在解码时额外转换像素格式.
   */
  enum wx_video_pixel_format target_pixel_format;
} wx_video_decoder_config_t;

#define WX_VIDEO_DECODER_LISTENER_TAG 0x40005

/**
 * @brief 视频流解码器回调
 *
 */
typedef struct wx_video_decoder_listener {
  struct wx_device common;

  /**
   * @brief 视频流解码器解码视频帧回调
   * 每当视频流解码得到一个视频帧，就应当尽快调用该函数.
   *
   * @param buffer 一帧视频帧数据内容
   * @param buffer_size 一帧视频帧数据大小
   */
  wx_error_t (*data)(const void* buffer, size_t buffer_size, struct wx_size);

  /**
   * @brief 通知视频流解码器在解码时出现的错误
   * @param encoder 出现错误的视频流解码器
   * @param user_data 通过 create_decoder 传入的 user_data
   * @param status 错误信息
   */
  void (*error)(struct wx_video_decoder* decoder,
                void* user_data,
                wx_error_t status);
} wx_video_decoder_listener_t;

#define WX_VIDEO_DECODER_TAG 0x40006

/**
 * @brief 视频流解码器
 *
 */
typedef struct wx_video_decoder {
  struct wx_device common;

  /**
   * @brief 更新视频流解码器数据
   * 视频流解码器接收到新一批数据后，对于每一个新解码得到的视频帧，应当调用
   * listener->data 回调通知解码后的数据.
   *
   * @param buffer 要解码的视频流的新一批数据
   * @param buffer_size 数据字节数
   */
  wx_error_t (*write)(struct wx_video_decoder* decoder,
                      const void* buffer,
                      size_t buffer_size);
} wx_video_decoder_t;

#define WX_VIDEO_STREAM_CONFIG_TAG 0x40007

typedef struct wx_video_stream_config {
  /**
   * @brief 要解码的视频流的原编码格式
   */
  enum wx_video_format format;

  /**
   * @brief 要解码的视频流的像素格式
   * 如果视频流帧中已经包含了像素格式, 该项可填 WX_VIDEO_PIXEL_FORMAT_UNKNOWN,
   * 由解码器自行判断.
   */
  enum wx_video_pixel_format pixel_format;

} wx_video_stream_config_t;

#define WX_VIDEO_STREAM_DATA_METADATA_TAG 0x40008

/**
 * @brief 视频流解码器配置
 *
 */
typedef struct wx_video_stream_data_metadata {
  /**
   * @brief 视频流像素宽度
   */
  uint32_t width;

  /**
   * @brief 视频流像素高度
   */
  uint32_t height;

  wx_video_rotation_t rotation;

} wx_video_stream_data_metadata_t;

#define WX_VIDEO_STREAM_TAG 0x40009

/**
 * @brief 视频输出流
 * 视频输出流可以输出指定编码的视频数据，解码由 wx_video_stream 内部自行实现
 *
 * 视频输出流需要绑定一个视频输出区域后才可显示. 参考
 * wx_surface::create_video_subsurface.
 */

typedef struct wx_video_stream {
  struct wx_struct common;

  /**
   * @brief 更新视频输出流数据
   * 若视频输出流为压缩流，解码需 wx_video_stream 自行实现
   *
   * @param buffer 视频帧
   * @param buffer_size 视频帧字节数
   */
  wx_error_t (*write)(struct wx_video_stream* stream,
                      const void* buffer,
                      size_t buffer_size,
                      const struct wx_video_stream_data_metadata* metadata);

  void (*close)(struct wx_video_stream*);
} wx_video_stream_t;

#define WX_VIDEO_MODULE_TAG 0x40000

typedef struct wx_video_module {
  struct wx_module common;

  /**
   * @brief 创建视频流编码器
   * @param config 编码配置
   * @param listener 回调
   * @param user_data 回调用户数据
   * @return 操作是否成功
   *   - WXERROR_NOT_FOUND: 不支持指定编码
   *   - WXERROR_INVALID_ARGUMENT: 配置不符合要求
   */
  wx_error_t (*create_encoder)(struct wx_video_module* module,
                               const struct wx_video_encoder_config* config,
                               struct wx_video_encoder_listener* listener,
                               void* user_data,
                               struct wx_video_encoder** encoder);

  /**
   * @brief 创建视频流解码器
   * @param config 解码配置
   * @param listener 回调
   * @param user_data 回调用户数据
   * @return 操作是否成功
   *   - WXERROR_NOT_FOUND: 不支持指定编码
   *   - WXERROR_INVALID_ARGUMENT: 配置不符合要求
   */
  wx_error_t (*create_decoder)(struct wx_video_module* module,
                               const struct wx_video_decoder_config* config,
                               struct wx_video_decoder_listener* listener,
                               void* user_data,
                               struct wx_video_decoder** decoder);

  /**
   * @brief 创建视频输出流
   */
  wx_error_t (*create_output_stream)(
      struct wx_video_module* module,
      const struct wx_video_stream_config* config,
      struct wx_video_stream** stream);
} wx_video_module_t;

WX_END_DECLS
