// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include "wmpf/macros.h"

typedef long wx_ssize_t;

/**
 * @brief 描述宿主结构体类型、大小及版本信息
 *
 * 如果你发现某个结构体最终的基类是 wx_struct, 那么这个结构体在创建时需要填写
 * wx_struct 的 tag、size、version, 以便 WMPF 及厂商接口在获取结构体对象
 * 时能保证收到的结构体是符合预期的.
 *
 * 比如 WMPF 获取了厂商提供的 wx_audio_device_out 对象, WMPF 为了校验厂商
 * 提供的对象是否确实是 wx_audio_device_out 类型, 防止运行时因为厂商传递了
 * 错误类型、错误大小、错误版本的结构体而崩溃, WMPF 会检查 wx_audio_device_out
 * 的 wx_struct 信息 (wx_audio_device_out::device::common::common).
 * 检查 .tag == WX_AUDIO_DEVICE_OUT_TAG,
 * .size == sizeof(wx_audio_device_out), .version == 0.
 * 检查通过后, 可以确保厂商确实传递了 wx_audio_device_out 对象, 而且双方使用
 * 的 wx_audio_device_out 对象的布局是一致的 (因为大小和版本号一致). WMPF
 * 可以直接使用这个对象, 而不会因为双方的对象布局不一致, 导致访问了错误的内存,
 * 导致崩溃.
 *
 * 注意: 我们强烈建议你在初始化包含 wx_struct 的结构体时, 调用 memset 将整个
 * 结构体 (例如 wx_audio_device_out) 置零后再填写内容, 否则若忘记填写某项时
 * (如 wx_audio_device_out::device::common::close), 由于内存默认不置 0,
 * 导致 WMPF 监测到 close 为非 0, 即厂商将 close 指向了可以关闭该音频设备的
 */
typedef struct wx_struct {
  /**
   * @brief 结构体类型
   * 具体值由使用 wx_struct 的结构体定义.
   * 用于验证结构体的类型, 如你获取了一个 wx_audio_device 对象, 你可以通过判断
   * .tag == WX_AUDIO_DEVICE_IN_TAG 来验证你获取到的是否是 wx_audio_device_in
   * 对象
   */
  uint32_t tag;

  /**
   * @brief 结构体大小
   * 构建结构体时需要传递最终结构体 (即没有任何类再继承自它) 的 sizeof 值,
   * 如传递 sizeof(wx_audio_device_in). 用于校验传递的结构体大小是否符合预期,
   * 避免 API 变更之后出现对不上结构体大小, 导致内存越界访问的问题.
   */
  uint32_t size;

  /**
   * @brief 结构体版本号
   * 代表一个结构体被修改了第 N 个版本.
   * 用于确保在调用方和被调用方的接口是对齐的.
   */
  uint32_t version;
} wx_struct_t;

/**
 * @brief 描述带长宽的大小
 *
 */
typedef struct wx_size {
  /**
   * @brief 宽度
   * 单位由使用方定义
   */
  size_t width;

  /**
   * @brief 高度
   * 单位由使用方定义
   */
  size_t height;
} wx_size_t;

/**
 * @brief 描述一个点
 *
 * 坐标系：左上方为原点，X 轴向右，Y 轴向下。
 *
 *  o --------> X 轴
 *  |
 *  |
 *  |
 *  v
 *
 *  Y 轴
 */
typedef struct wx_point {
  /**
   * @brief 点的横坐标
   * 单位由使用方定义
   */
  size_t x;

  /**
   * @brief 点的纵坐标
   * 单位由使用方定义
   */
  size_t y;
} wx_point_t;

/**
 * @brief 描述一个区域/矩形
 *
 */
typedef struct wx_region {
  /**
   * @brief 矩形的横坐标
   * 单位由使用方定义
   */
  size_t x;

  /**
   * @brief 矩形的纵坐标
   * 单位由使用方定义
   */
  size_t y;

  /**
   * @brief 矩形的宽度
   * 单位由使用方定义
   */
  size_t width;

  /**
   * @brief 矩形的高度
   * 单位由使用方定义
   */
  size_t height;
} wx_region_t;
