#pragma once

#include <stdbool.h>

#include "wmpf/macros.h"
#include "wmpf/operation.h"

WX_BEGIN_DECLS

/**
 * @brief 检测设备是否已经被注册
 * 需要在 wx_init 调用完成之后才能调用本函数.
 *
 * @param is_registered_out 输出设备注册情况.
 * @return 检测过程中是否出现错误, 比如数据文件夹不可读写, 或者 rpmb
 * 设备不能被正常访问.
 *   - WXERROR_FAILED_PRECONDITION: wx_init 未被调用
 *   - 其他错误: 可能是访问 data_dir 或者 rpmb 设备失败.
 */
WX_API wx_error_t wx_device_is_registered(bool* is_registered_out);

/**
 * @brief 注册设备
 * 需要在 wx_init 调用完成之后才能调用本函数.
 *
 * 设备票据是微信后台向当前设备发放的访问小程序 VoIP 相关接口的唯一凭证,
 * 具有一次性获取、本地永久储存的特点. 在初始化 WMPF 之前, 你需要申请
 * Model ID, 并为设备分配设备 ID, 并通过调用微信后台的一次性接口为当前
 * 设备获取设备票据 (即每台设备仅能调用一次微信后台接口获取当前设备的设备
 * 票据), 并把设备票据保存在可靠的存储区内 (由于后台只能获取一次当前设备
 * 的), 同时确保该信息不会泄漏.
 *
 * 该函数是异步函数, 会产生网络请求, 请确保在调用本函数时网络通畅.
 *
 * 该函数会存储设备 ID、设备注册信息、设备证书. 如果其中有任意信息丢失,
 * 该函数都需要被再次调用以恢复信息. 但若该函数再次调用之后返回
 * WXERROR_ALREADY_EXISTS, 你可能需要更换设备 ID.
 *
 * 建议你在调用 wx_device_is_registered 查询到设备未被注册后再调用本函数,
 * 因为如果设备已经被注册过, 你就不需要再获取 SNTicket 了.
 * 如果你在没有调用 wx_device_is_registered 函数查询注册状态, 且当前设备
 * 已经注册过了, 该函数会返回 WXERROR_PERMISSION_DENIED.
 *
 * 一般情况下，你可以参考以下代码调用本函数:
 *
 * @code {.c}
 * bool is_registered = false;
 * wx_error_t is_reg_ret = wx_device_is_registered(&is_registered);
 * if (is_reg_ret != WXERROR_OK) {
 *   // 你需要在这里处理错误码.
 * } else if (!is_registered) {
 *   const char* sn_ticket = "";
 *   wx_operation_t op = wx_device_register(sn_ticket);
 *   wx_error_t ret = wx_operation_wait(op, 0);
 *   if (ret != WX_ERROR_OK) {
 *     // 你需要在这里处理错误码.
 *   } else {
 *     // 设备注册成功.
 *   }
 * }
 * @endcode
 *
 * 注意，如果数据文件夹 data_dir 可写入的磁盘空间不足, 或者没有写入权限,
 * 将会导致设备注册失败.
 *
 * @param sn_ticket 设备票据
 * @return 注册是否成功
 *   - WXERROR_INVALID_ARGUMENT: SNTicket 不合法或超时.
 *   - WXERROR_ALREADY_EXISTS: 当前设备已经注册过了, 不需要再调用本函数.
 *   - WXERROR_PERMISSION_DENIED: 该设备 ID 已经在其他时刻被注册过了
 *   - WXERROR_RESOURCE_EXHAUSTED: 网络不连通
 *   - WXERROR_DATA_LOSS: 保存设备注册信息失败
 *   - WXERROR_FAILED_PRECONDITION: wx_init 未被调用; ModelID 不合法.
 * @see
 * https://developers.weixin.qq.com/miniprogram/dev/OpenApiDoc/hardware-device/getSnTicket.html
 */
WX_API wx_operation_t wx_device_register(const char* sn_ticket);

WX_END_DECLS
