// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include "wmpf/error.h"
#include "wmpf/macros.h"
#include "wmpf/operation.h"
#include "wmpf/types.h"

WX_BEGIN_DECLS

typedef struct wx_voip_session* wx_voip_session_t;

typedef enum wx_voip_session_type {
  WX_VOIP_SESSION_VIDEO = 0,  // 音视频通话
  WX_VOIP_SESSION_AUDIO = 1,  // 纯音频通话
} wx_voip_session_type_t;

typedef enum wx_voip_session_scene {
  WX_VOIP_SESSION_SCENE_IOT = 0,     // IoT VoIP 通话模式
  WX_VOIP_SESSION_SCENE_SCHOOL = 1,  // 校园 VoIP 通话模式
} wx_voip_session_scene_t;

typedef enum wx_voip_session_status {
  WX_VOIP_SESSION_IDLE = 0,     // 初始状态
  WX_VOIP_SESSION_CALLING = 1,  // 拨打电话中
  WX_VOIP_SESSION_TALKING = 2,  // 通话中 (被拨打的微信用户接听了电话)
  WX_VOIP_SESSION_REJECTED = 3,  // 被拨打的微信用户拒绝接听电话
  WX_VOIP_SESSION_CANCELED = 4,  // 拨打过程中, 设备取消了电话拨打
  WX_VOIP_SESSION_HANGUP_BY_CALLER = 5,  // 通话时设备挂断了电话
  WX_VOIP_SESSION_HANGUP_BY_CALLEE = 6,  // 通话时被拨打的微信用户挂断了电话
  WX_VOIP_SESSION_ABORTED = 7,  // 发生异常
  WX_VOIP_SESSION_BUSY = 8,     // 被拨打的微信用户处于占线状态
  WX_VOIP_SESSION_TIMEOUT = 9,  // 超时未接听
} wx_voip_session_status_t;

typedef enum wx_voip_hangup_reason {
  WX_VOIP_HANGUP_REASON_UNKNOWN = 0,
  WX_VOIP_HANGUP_REASON_MANUAL = 1,  // 用户手动挂断/取消/拒绝通话

  // 主叫挂断原因
  WX_VOIP_HANGUP_REASON_SYSTEM = 6,  // 被系统电话挂断
  WX_VOIP_HANGUP_REASON_APP = 7,     // 被其他应用挂断
  WX_VOIP_HANGUP_REASON_DEVICE = 8,  // 采集播放设备启动失败

  WX_VOIP_HANGUP_REASON_TIMEOUT = 10, // 超时挂断
  WX_VOIP_HANGUP_REASON_REJECT = 11, // 拒绝通话，指的是没有进入通话即挂断
} wx_voip_hangup_reason_t;

typedef enum wx_voip_member_camera_status {
  WX_VOIP_MEMBER_CAMERA_STATUS_OPEN = 0,   // 该成员摄像头开启
  WX_VOIP_MEMBER_CAMERA_STATUS_CLOSE = 1,  // 该成员摄像头关闭
} wx_voip_member_camera_status_t;

/**
 * @brief 表示一个 VoIP 角色, 可以是拨打方, 也可以是被拨打方.
 *
 */
typedef struct wx_voip_member {
  /**
   * @brief
   * 对于设备：
   * IoT VoIP 模式下该项填 "" (该项由厂商小程序调用
   * `wx.requestDeviceVoIP` 时设置, 这里必须填空字符串,
   * 由微信后端替换成实际设备名称).
   *
   * 校园模式下该项填学生名称.
   *
   * 对于被拨打的微信用户：
   * 该项填微信用户昵称（显示在通话页面中，可以填房间号之类的名称）
   */
  const char* name;

  /**
   * @brief
   * 对于设备：
   * IoT VoIP 模式下该项填 设备 ID (SN).
   * 校园模式下该项填学生 ID.
   *
   * 对于被拨打的微信用户:
   * 微信用户的 OpenID.
   * OpenID 需要在微信手机客户端小程序侧通过 `wx.login` 获取,
   * 并通过你自建的某种通信渠道传递到设备上后，填到此项.
   */
  const char* id;

  /**
   * @brief 表示该 VoIP 通话成员是否开启摄像头
   * 对于设备：
   * 如果设备不具备摄像头，该项填 WX_VOIP_MEMBER_CAMERA_STATUS_CLOSE.
   * 如果设备具备摄像头，该项填 WX_VOIP_MEMBER_CAMERA_STATUS_OPEN.
   *
   * 对于被拨打的微信用户：
   * 如果设备不具备屏幕，或者不需要查看被拨打的微信用户的摄像头内容，该项填
   * WX_VOIP_MEMBER_CAMERA_STATUS_CLOSE.
   * 如果设备具备屏幕，且需要查看被拨打的微信用户的摄像头内容，该项填
   * WX_VOIP_MEMBER_CAMERA_STATUS_OPEN.
   */
  wx_voip_member_camera_status_t camera_status;
} wx_voip_member_t;

/**
 * @brief VoIP 会话的额外配置
 * 使用本结构体前, 建议你对这个结构体清零后使用
 * 否则当 API 变更时, 你的程序无法观测到本结构体变化, 导致崩溃。
 */
typedef struct wx_voip_session_config {
  /**
   * @brief 摄像头像素宽度
   * 传 0 为默认配置.
   *
   * 这里需要传旋转之前的摄像头图像的像素宽度.
   * 即如果你修改了 camera_rotation, 该项不需要修改 (如交换宽高).
   */
  uint32_t camera_width;

  /**
   * @brief 摄像头像素宽度
   * 传 0 为默认配置.
   *
   * 这里需要传旋转之前的摄像头图像的像素高度.
   * 即如果你修改了 camera_rotation, 该项不需要修改 (如交换宽高).
   */
  uint32_t camera_height;

  /**
   * @brief 摄像头图像旋转角度
   *
   * 只能填 0, 90, 180, 270.
   */
  uint32_t camera_rotation;

  /**
   * @brief 小程序页面自定义参数
   *
   * 接听方点击通知打开小程序时，会作为 query 拼接到插件页面路径后。 格式如
   * a=1&b=2 ; 可在接听端小程序内通过 getPluginEnterOptions 接口获取到
   * 注意: 该项需要厂商自行转义.
   */
  const char* custom_query;
} wx_voip_session_config_t;

#define WX_VOIP_SESSION_LISTENER_TAG 0xF00001

typedef struct wx_voip_session_listener {
  wx_struct_t common;

  /**
   * @brief 当前 VoIP 通话状态改变
   *
   */
  void (*status)(wx_voip_session_t session,
                 void* user_data,
                 wx_voip_session_status_t);

} wx_voip_session_listener_t;

enum wx_voip_delayed_loading_state {
  WX_VOIP_DELAYED_LOADING_STATE_DISABLED,
  WX_VOIP_DELAYED_LOADING_STATE_ENABLED,
  WX_VOIP_DELAYED_LOADING_STATE_READY,
};

/**
 * @brief 设置部分 so 异步加载
 *
 * 背景:
 *   由于 wxvoip 库总体积比较大，通常大于 10M，难以在一些内存比较小、拨打 VoIP
 *   时需临时启动系统的设备上将库文件打包进 init.rd 快速加载进系统内，而需要将库
 *   文件放置于外置 Flash 等加载速度比较慢的存储区域内，可能会造成在 wxvoip 库
 *   本身 wx_init() 函数调用的时机被严重推迟，导致拨打 VoIP 时，对方手机微信收
 *   到铃声提醒的耗时比较大。
 * 解决思路:
 *   某些设备上，由于机器性能原因，慢速存储区域的加载时间可达 5s, wx_init 本身
 *   耗时可达 6s, 若这里能够并行执行, 能将总时间 5s+6s=11s 压缩至
 *   max(5s,6s)=6s 完成.
 * 解决方案:
 *   由于被拨打的用户在收到 VoIP 强提醒铃声时，往往需要数秒以完成掏出手机、打开
 *   微信应用、点击接听电话按钮、打开并加载厂商小程序等步骤，才能正式接听电话，这
 *   里的等待时间其实也有数秒的时间。因此我们这里允许将发送给手机微信的铃声提醒过
 *   程和加载 wxvoip 库其他 so 的过程并行执行，即厂商将 wxvoip 库放置于
 *   init.rd 内, 剩余的 so 继续放置于外置 Flash 等慢速存储区域内。系统加载完成
 *   init.rd 后, 立刻启动慢速存储区域的初始化（如设备启动、文件系统初始化等操
 *   作），此同时，调用 wx_init 和 wx_voip_session_call 等函数先发送 VoIP
 *   强提醒铃声。等待到慢速存储区域加载完成之后，通过本函数通知 wxvoip 库可以继
 *   续执行剩余库文件的逻辑，正式发起 VoIP 通话。
 * 使用方法：
 *   在 wx_init() *之前* 调用
 *   wx_voip_delay_loading(WX_VOIP_DELAYED_LOADING_STATE_ENABLE)
 *   以启用如下 so 的异步加载:
 *   - libilink2.so (如有)
 *   - libilink_network.so
 *   - libilink_protobuf.so
 *   - libconfService.so
 *   - libvoipChannel.so
 *   - libvoipCodec.so
 *   - libvoipComm.so
 *   以下 so 需要在系统启动后、发送 VoIP 强提醒铃声前立刻准备好：
 *   - libwxvoip.so
 *   - libowl.so
 *   - libcrash.so (如有)
 *   被异步加载的 so 可以被放到慢速存储区域内，与 wx_init 执行的同时并行加载。
 *   在慢速存储区域准备好后，调用
 *   wx_voip_delay_loading(WX_VOIP_DELAYED_LOADING_STATE_READY)
 *   通知 wxvoip 可以完成剩余库文件的逻辑，以正式发起 VoIP 通话.
 *   需要注意, delay_loading(READY) 函数内会执行 dlopen 上述 so 的操作,
 *   因此该函数会阻塞执行直到上述 so 均被 dlopen 完成.
 * @note
 *   如果需要调用 wx_voip_delay_loading，你只能在第一次 wx_init() 之前
 *   调用 enable, 在慢速存储加载完成之后调用 ready. 不能在调用 ready 之后再
 *   调用 enable；也不能 enable 之后再 enable.
 *   因此如果存在多次调用 wx_init() 和 wx_stop() 的情况，请注意 ENABLE 调用
 *   只能在第一次 wx_init() 之前被调用。READY 只能在慢速存储加载完成后调用一次。
 * @return
 *   - WXERROR_FAILED_PRECONDITION: 在 wx_init() 之后调用 enable，或者多次
 *                                  ready.
 */
WX_API wx_error_t
wx_voip_delay_loading(enum wx_voip_delayed_loading_state state);

/**
 * @brief 创建 VoIP 会话对象
 *
 * @param listener 回调
 * @param user_data 回调用户数据
 * @param self_member 本设备的 VoIP 角色设置
 * @param session_out VoIP 会话结果
 * @return
 *   - WXERROR_INVALID_ARGUMENT: 参数错误
 *   - WXERROR_FAILED_PRECONDITION: wx_init 未调用或未完成.
 */
WX_API wx_error_t
wx_voip_session_new(wx_voip_session_type_t,
                    wx_voip_session_scene_t,
                    const struct wx_voip_session_listener* listener,
                    void* user_data,
                    const wx_voip_member_t* self_member,
                    const wx_voip_session_config_t* config,
                    wx_voip_session_t* session_out);

/**
 * @brief 拨打 VoIP 通话给指定的微信用户
 *
 * 厂商需要自行实现用户的通讯录功能.
 * 厂商需要在自己的小程序内提示用户向指定的设备 (设备 ID) 授权 VoIP 通话权限后,
 * 方可在携带该设备 ID 的设备上向该用户拨打 VoIP 通话.
 *
 * 该函数是异步函数, 会在成功发起通话请求之后异步返回, 如果你要检测后续通话状态,
 * 你应该通过 wx_voip_session_listener 来监听本次通话之后是否被接听或被对方挂断.
 *
 * @param name 微信用户昵称（显示在通话页面中，可以填房间号之类的名称）
 * @param id 微信用户的 OpenID.
 * @return 该函数是异步操作, 你需要调用 wx_operation_await 或者
 *   wx_operation_wait 等待操作结束, 否则会导致内存泄漏.
 *   - WXERROR_FAILED_PRECONDITION: 当前正处于其他通话中.
 * 请挂断之后再调用本函数.
 *   - WXERROR_RESOURCE_EXHAUSTED: 当前无网络
 *   - WXERROR_INVALID_ARGUMENT: openid/snticket 错误
 *   - WXERROR_PERMISSION_DENIED: 被拨打的用户未向当前设备授权拨打 VoIP
 * 通话权限. 或者当前小程序没有拨打 VoIP 通话权限.
 *   - WXERROR_INTERNAL: 内部错误
 *   - WXERROR_UNAVAILABLE: 当前微信小程序没有拨打 VoIP 通话权限,
 * 或未购买通话时长.
 */
WX_API wx_operation_t wx_voip_session_call(wx_voip_session_t session,
                                           const wx_voip_member_t* callee);

/**
 * @brief 得到当前 session 的 roomid
 
 *  roomid，是当前 voip 通话所在房间的标识。
 * 
 * @param session
 * @return roomid 字符串 或 "";
 */
WX_API const char* wx_voip_session_get_roomid(wx_voip_session_t session);

/**
 * @brief 根据 roomid 加入通话
 *
 *  小程序呼叫 Linux 设备时，roomid 流转到设备后，设备通过此接口加入通话
 * 
 * @param session
 * @param roomid 插件向设备发起通话后，roomid 为此通话标识，设备 sdk 使用它加入通话
 * @return roomid 字符串 或 NULL。
 */
WX_API wx_operation_t wx_voip_listener_join(wx_voip_session_t session,
                                           const char *roomid);

/**
 * @brief 挂断 VoIP 通话
 *
 * 该函数需要在用户手动挂断, 或者因为系统的其他原因需要挂断 VoIP 通话时调用.
 *
 * @param reason 挂断 VoIP 通话的原因
 * @return 操作是否成功
 */
WX_API wx_operation_t wx_voip_session_hangup(wx_voip_session_t session,
                                             wx_voip_hangup_reason_t reason);

/**
 * @brief 小程序呼叫设备时，设备端作为接听者, 在未创建 session 时可直接挂断。
 *   注意：如果已创建 session，则需要使用 wx_voip_session_hangup 进行挂断。
 * @param reason 挂断 VoIP 通话的原因
 * @return 操作是否成功
 */
WX_API wx_operation_t wx_voip_listener_hangup(const char *roomid,
                                             wx_voip_hangup_reason_t reason);

/**
 * @brief 销毁 VoIP 会话对象
 *
 * @param session
 * @return WX_API
 */
WX_API void wx_voip_session_destroy(wx_voip_session_t session);

WX_END_DECLS
